#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <unistd.h>

#include "main.h"
#include "feature.h"
#include <debug.h> /* from libmej */
#include "debug.h"
#include "mem.h"
#include "options.h"
#ifdef PIXMAP_SUPPORT
# include "pixmap.h"
#  include "screen.h"
# ifdef POSIX_THREADS
#  include <pthread.h>
#  include "threads.h"
 # endif
 #endif
 #ifdef PIXMAP_SCROLLBAR
 # include "scrollbar.h"
 #endif

 #ifdef USE_IMLIB
 # include "eterm_imlib.h"
 #endif

 #ifdef PIXMAP_SUPPORT
 /* Specifying a single extension is irrelevant with Imlib. -vendu */
 # ifdef USE_IMLIB
 #  define PIXMAP_EXT NULL
 # endif

 extern XWindowAttributes attr;
 extern char * rs_path;
 extern const char * rs_pixmapScale;
 # ifdef PIXMAP_OFFSET
 extern const char * rs_pixmapTrans;
 # endif

 extern short bg_needs_update;

 /* the originally loaded pixmap and its scaling */
 void set_bgPixmap (const char * /* file */);
 void resize_subwindows(int, int);
 # ifdef POSIX_THREADS
 #  ifdef MUTEX_SYNCH
 extern pthread_mutex_t mutex;
 #  endif
 extern short bg_set;
 extern char *sig_to_str(int sig);
 extern pthread_t resize_sub_thr;
 extern pthread_attr_t resize_sub_thr_attr;
 # endif

 # ifdef PIXMAP_OFFSET
 extern XSizeHints szHint;
 Pixmap offset_pixmap(Pixmap, int, int);
 # endif

 pixmap_t bgPixmap = { 0, 0, 50, 50, None };
 # ifdef PIXMAP_SCROLLBAR
 pixmap_t upPixmap = { 0, 0, 50, 50, None };
 pixmap_t dnPixmap = { 0, 0, 50, 50, None };
 pixmap_t sbPixmap = { 0, 0, 50, 50, None };
 pixmap_t saPixmap = { 0, 0, 50, 50, None };
 # endif   

 # ifdef USE_IMLIB
 ImlibData *imlib_id = NULL;
 imlib_t imlib_bg = { NULL, 0, 0 };
 #  ifdef PIXMAP_SCROLLBAR
 imlib_t imlib_up = { NULL, 0, 0 };
 imlib_t imlib_dn = { NULL, 0, 0 };
 imlib_t imlib_sb = { NULL, 0, 0 };
 imlib_t imlib_sa = { NULL, 0, 0 };
 #  endif

 Image
  *ReadImageViaImlib(Display *d, char *filename)
 {
	 Image *tmp;

	 D_IMLIB2("ReadImageViaImlib(%s)\n", filename);

	 if (imlib_id == NULL) {
 #  ifdef NEW_IMLIB
	 D_IMLIB("Imlib_init()\n");
 #  else
	 D_IMLIB("ImlibInit()\n");	
 #  endif
	 imlib_id = ImlibInit(d);
	 }
 #  ifdef NEW_IMLIB
	 D_IMLIB2("Imlib_load_image(%s)\n", filename);
 #  else
 #   ifdef OLD_IMLIB
	 D_IMLIB2("ImlibLoadImage(%s)\n", filename);
 #   endif
 #  endif
	 tmp = ImlibLoadImage(imlib_id, filename, NULL);
	 return tmp;
 }

 #endif /* USE_IMLIB */

 /*
  * These GEOM strings indicate absolute size/position:
  * @ `WxH+X+Y'
  * @ `WxH+X'	-> Y = X
  * @ `WxH'	-> Y = X = 50
  * @ `W+X+Y'	-> H = W
  * @ `W+X'	-> H = W, Y = X
  * @ `W'	-> H = W, X = Y = 50
  * @ `0xH'	-> H *= H/100, X = Y = 50 (W unchanged)
  * @ `Wx0'	-> W *= W/100, X = Y = 50 (H unchanged)
  * @ `=+X+Y'	-> (H, W unchanged)
  * @ `=+X'	-> Y = X (H, W unchanged)
  *
  * These GEOM strings adjust position relative to current position:
  * @ `+X+Y'
  * @ `+X'	-> Y = X
  *
  * And this GEOM string is for querying current scale/position:
  * @ `?'
  */


 /*   '[', 2*4 + 2*3 digits + 3 delimiters, ']'. -vendu */
 # define GEOM_LEN 19    /* #undef'd immediately after scale_pixmap(). - vendu */
 int
 scale_pixmap (const char * geom, pixmap_t pmap)
 {
 # if 0
	 /* should be big enough */
	 /*   static char str [] = "[1000x1000+100+100]"; */
 # endif
	  static char str[GEOM_LEN+1] = {'\0'};
	 int w = 0, h = 0, x = 0, y = 0;
	 int flags;
	 int changed = 0;
	 char *p;
	 int n;

	 if (geom == NULL) return 0;

 #if 0
 # ifdef DEBUG_PIXMAP    
	 fprintf(stderr, "scale_pixmap(\"%s\")\n", geom);
 # endif
	 if (!strcmp (geom, "?")) {
	 sprintf (str,
		  "[%dx%d+%d+%d]",
		  bgPixmap.w,
		  bgPixmap.h,
		  bgPixmap.x,
		  bgPixmap.y);
	 xterm_seq (XTerm_title, str);
	 return 0;
	 }
 #endif

	 if ((p = strchr (geom, ';')) == NULL)
	   p = strchr (geom, '\0');
	 n = (p - geom);
	 if (n > GEOM_LEN - 1) return 0;

	 strncpy (str, geom, n);
	 str [n] = '\0';

	 flags = XParseGeometry (str, &x, &y, &w, &h);

	 if (!flags) {
	 flags |= WidthValue;    /* default is tile */
	 w = 0; 
	 }

	 if (flags & WidthValue) {
	 if (!(flags & XValue)) { x = 50; }
	 if (!(flags & HeightValue)) h = w;

	 if (w && !h) {
		 w = pmap.w * ((float)w/100);
		 h = pmap.h;
	 } else if (h && !w) {
		 w = pmap.w;
		 h = pmap.h * ((float)h/100);
	 }

	 /* FIXME: Should be fixed to use automagic limits. -vendu */

	 if (w > 1000) w = 1000;
	 if (h > 1000) h = 1000;

	 if (pmap.w != w) {
		 pmap.w = w;
		 changed++;
	 }
	 if (pmap.h != h) {
		 pmap.h = h;
		 changed++;
	 }
	 }

	 if (!(flags & YValue)) {
	 if (flags & XNegative) flags |= YNegative;
	 y = x;
	 }
	 if (!(flags & WidthValue) && geom [0] != '=') {
	 x += pmap.x;
	 y += pmap.y;
	 } else {
	 if (flags & XNegative) x += 100;
	 if (flags & YNegative) y += 100;
	 }

	 x = (x <= 0 ? 0 : (x >= 100 ? 100 : x));
	 y = (y <= 0 ? 0 : (y >= 100 ? 100 : y));;
	 if (pmap.x != x) {
	 pmap.x = x;
	 changed++;
	 }
	 if (pmap.y != y) {
	 pmap.y = y;
	 changed++;
	 }
	 return changed;
 }
 #  undef GEOM_LEN

 void
 resize_pixmap (Window win, imlib_t image, pixmap_t pmap,
			int which, int enable_offset)
 {
	 XGCValues gcvalue;
	 GC gc;
	 unsigned int width = 0;
	 unsigned int height = 0;
	 float p, incr;
	 int xsize, ysize;
	 Pixmap pixmap = None;
 #  ifdef PIXMAP_OFFSET
	 int x_offset, y_offset;
	 static int rendered = 0;
	 Window W_dummy;
 #   if 0
	 int dummy;
	 unsigned int u_dummy;
	 long l_dummy;
 #   endif
 #  endif

	 if (!image.im) {
	 D_PIXMAP("resize_pixmap(): no image loaded\n");

	 if ((Options & Opt_borderless) && !(background_is_pixmap())) {
		 XSetWindowBackground(Xdisplay, win, PixColors[bgColor]);
		 XClearWindow(Xdisplay, win);
	 }

	 return;
	 }

	 switch(which) {
	  case pixmap_bg:
	 width = TermWin_TotalWidth ();
	 height = TermWin_TotalHeight ();
	 break;
	  default:
		break;
    }
    
    if (!(width) || !(height))
      return;
    
    /*    D_PIXMAP("resize_pixmap()\n"); */
    
    if (enable_offset && (Options & Opt_pixmapTrans)) {
	XTranslateCoordinates(Xdisplay, win, Xroot,
			      0, 0, &x_offset, &y_offset, &W_dummy);	
    }
    
    xsize = image.im->rgb_width;
    ysize = image.im->rgb_height;
    
    gcvalue.foreground = PixColors [bgColor];
    gc = XCreateGC (Xdisplay, win, GCForeground, &gcvalue);
    
    if (image.im) {
	int w = pmap.w;
	int h = pmap.h;
	int x = pmap.x;
	int y = pmap.y;
	
	/*
	 * don't zoom pixmap too much nor expand really small pixmaps
	 */
	
	/* FIXME: automagic limits. -vendu */
	if (w > 1000 || h > 1000) {
	    w = 1;
	} else if (width > (10 * xsize) || height > (10 * ysize))
	      w = 0;		/* tile */
	
	if (w) {
	    
	    /*
	     * horizontal scaling
	     */
	    
	    D_PIXMAP("resize_pixmap(): horizontal scaling\n");
	    
	    incr = (float)xsize;
	    
	    p = 0;
	    
	    if (w == 1) {
		/* display image directly - no scaling at all */
		incr = width;
		
		D_PIXMAP("resize_pixmap(): no horizontal scaling\n");
		
		if (xsize <= width) {
		    w = xsize;
		    x = (width - w) / 2;
		    w += x;
		} else {
		    x = 0;
		    w = width;
		}
	    } else if (w < 10) {
		incr *= w;	/* fit W images across screen */
		x = 0;
		w = width;
	    } else {
		incr *= 100.0 / w;
		/* contract */ 
		if (w < 100) {
		    w = (w * width) / 100;
		    /* position */
		    if (x >= 0)	{
			float pos;
			pos = (float)x / 100 * width - (w/2);
			
			x = (width - w);
			if (pos <= 0) x = 0; else if (pos < x) x = pos;
		    } else {
			x = (width - w) / 2;
		    }
		    w += x;
		} else if (w >= 100) {
		    /* expand
		     */
		    /* position */
		    if (x > 0) {
			float pos;
			pos = (float)x / 100 * xsize - (incr/2);
			
			p = xsize - (incr);
			if (pos <= 0) p = 0; else if (pos < p) p = pos;
		    }
		    x = 0;
		    w = width;
		}
	    }
	    incr /= width;
	    
	    /*
	     * vertical scaling
	     */
	    
	    D_PIXMAP("resize_pixmap(): vertical scaling\n");
	    
	    incr = (float)ysize;
	    p = 0;
	    
	    if (h == 1) {
		/* display image directly - no scaling at all */
		incr = height;
		
		D_PIXMAP("resize_pixmap(): no vertical scaling\n");
		
		if (ysize <= height) {
		    h = ysize;
		    y = (height - h) / 2;
		    h += y;
		} else {
		    y = 0;
		    h = height;
		}
	    } else if (h < 10) {
		incr *= h;	/* fit H images across screen */
		y = 0;
		h = height;
	    } else {
		incr *= 100.0 / h;
		/* contract */
		if (h < 100) {
		    h = (h * height) / 100;
		    /* position */
		    if (y >= 0) {
			float pos;
			pos = (float)y / 100 * height - (h/2);
			
			y = (height - h);
			if (pos < 0.0f) y = 0; else if (pos < y) y = pos;
		    } else {
			y = (height - h) / 2;
		    }
		    h += y;
		} else if (h >= 100) {
		    /* expand
		     */
		    /* position */
		    if (y > 0) {
			float pos;
			pos = (float)y / 100 * ysize - (incr/2);
			
			p = ysize - (incr);
			if (pos < 0) p = 0; else if (pos < p) p = pos;
		    }
		    y = 0;
		    h = height;
		}
	    }
	    incr /= height;
	    
	    image.last_w = w; image.last_h = h;
	    
	    fprintf(stderr, "ImlibRender(%dx%d)\n", w, h);
	    ImlibRender(imlib_id, image.im, w, h);
	    
	    D_IMLIB("pmap.pixmap = ImlibCopyImageToPixmap()\n");
	    pmap.pixmap = ImlibCopyImageToPixmap(imlib_id, image.im);
	    
	    D_IMLIB("ImlibDestroyImage()\n");
	    ImlibDestroyImage(imlib_id, image.im);
	    image.im = NULL;
	    
	} else {
	    /* if (w), light years above. -vendu */
	    /* tiled */
	    
	    D_PIXMAP("resize_pixmap(): tiling pixmap\n");
	    
	    if (!((Options & Opt_pixmapTrans) && rendered)) {
		D_IMLIB3("ImlibRender(%dx%d)\n", xsize, ysize);
		ImlibRender(imlib_id, image.im, xsize, ysize);
		rendered = 1;
	    }
	    
	    D_IMLIB("pixmap = ImlibCopyImageToPixmap()\n");
	    pixmap = ImlibCopyImageToPixmap(imlib_id, image.im);
	    
	    D_IMLIB("ImlibDestroyImage()\n");
	    ImlibDestroyImage(imlib_id, image.im);
	    image.im = NULL;
	    
#  ifdef PIXMAP_OFFSET	    
	    if (Options & Opt_pixmapTrans) {
#   ifdef DEBUG_PIXMAP
		fprintf(stderr, "pmap.pixmap = offset_pixmap(pixmap)\n",
			xsize, ysize);
#   endif
		pmap.pixmap = offset_pixmap(pixmap, x_offset, y_offset);
		
		D_IMLIB("ImlibFreePixmap(pixmap)\n");
		ImlibFreePixmap(imlib_id, pixmap);
		pixmap = None;
	    } else
#  endif   
	      {
#  ifdef DEBUG_PIXMAP
		  fprintf(stderr, "XCreatePixmap(pmap.pixmap(%d,%d))\n",
			  width, height);
#  endif
		  pmap.pixmap = XCreatePixmap(Xdisplay, win,
					      width, height, Xdepth);
		  
		  for (y = 0; y < height; y += ysize) {
		      unsigned int h = (height - y);
		      
		      if (h > ysize) h = ysize;
		      
		      for (x = 0; x < width; x += xsize) {
			  unsigned int w = (width - x);
			  
			  if (w > xsize) w = xsize;
#  ifdef DEBUG_PIXMAP
			  fprintf(stderr,
				  "XCopyArea(pixmap(%dx%d)->pmap.pixmap)\n",
				  w, h);
#  endif			  
			  XCopyArea (Xdisplay,
				     pixmap,
				     pmap.pixmap,
				     gc,
				     0, 0, w, h, x, y);
		      }
		  }
		  
		  D_IMLIB("ImlibFreePixmap(pixmap)\n");
		  ImlibFreePixmap(imlib_id, pixmap);
		  pixmap = None;
	      }
	}
    } else
      XFillRectangle (Xdisplay, pixmap, gc,
		      0, 0, width, height);
    
    /* PIXMAP_BUFFERING isn't supported (yet?), so this is thoroughly broken
     * at the moment >:) -vendu
     */
#  ifdef PIXMAP_BUFFERING
    if (TermWin.buf_pixmap) XFreePixmap (Xdisplay, TermWin.buf_pixmap);
    TermWin.buf_pixmap = XCreatePixmap (Xdisplay, TermWin.vt,
					width, height, Xdepth);
    XCopyArea (Xdisplay,
	       TermWin.pixmap,
	       TermWin.buf_pixmap,
	       gc,
	       0, 0,
	       width, height,
	       0, 0);
    XSetWindowBackgroundPixmap (Xdisplay, TermWin.vt, TermWin.buf_pixmap);
#  else	/* PIXMAP_BUFFERING */
    if (pmap.pixmap != None) {
	D_PIXMAP("XSetWindowBackgroundPixmap(pmap.pixmap)\n");
	XSetWindowBackgroundPixmap (Xdisplay, win, pmap.pixmap);
	D_IMLIB("ImlibFreePixmap(pmap.pixmap)\n");
	ImlibFreePixmap(imlib_id, pmap.pixmap);
	pmap.pixmap = None;
#  endif	/* PIXMAP_BUFFERING */
	XSync (Xdisplay, 0);
	XFreeGC (Xdisplay, gc);
	
	XClearWindow (Xdisplay, win);
	XFlush (Xdisplay);
	
	XSync (Xdisplay, 0);
    }
}
# else	/* PIXMAP_SUPPORT */
#  define scale_pixmap(str,p) ((void)0)
#  define resize_pixmap() ((void)0)
# endif	/* PIXMAP_SUPPORT */

# if defined(PIXMAP_SUPPORT) || (MENUBAR_MAX)
/*
 * search for FILE in the current working directory, and within the
 * colon-delimited PATHLIST, adding the file extension EXT if required.
 *
 * FILE is either @ or zero terminated
 */
const char *
search_path(const char *pathlist, const char *file, const char *ext)
{
    /* FIXME: the 256 below should be changed to some #define in <dirent.h> */
    static char name [256];
    char * p;
    const char * path;
    int maxpath, len;
    struct stat fst;
    
#  ifdef DEBUG_OPTIONS
    D_OPTIONS4("search_path(\"%s\", \"%s\", \"%s\") called.\n",
	     pathlist, file, ext);
    D_OPTIONS2("search_path():  Checking for file \"%s\"\n", file);
#  endif
    if (!access(file, R_OK)) {
	if (stat(file, &fst)) {
	    dprintf(("Unable to stat %s -- %s\n", file, strerror(errno)));
	} else {
	    dprintf(("Stat returned mode 0x%08o, S_ISDIR() == %d\n", fst.st_mode, S_ISDIR(fst.st_mode)));
	}
	if (!S_ISDIR(fst.st_mode)) return file;
    }
    
    /* Changed to use '@' as the delimiter. -vendu */
    if ((p = strchr(file, '@')) == NULL) p = strchr(file, '\0');
    len = (p - file);
    
    /* check about adding a trailing extension */
    if (ext != NULL) {
	
	char *dot;
	
	dot  = strrchr(p, '.');
	path = strrchr(p, '/');
	if (dot != NULL || (path != NULL && dot <= path)) ext = NULL;
    }
    
    /* leave room for an extra '/' and trailing '\0' */
    maxpath = sizeof(name) - (len + (ext ? strlen(ext) : 0) + 2);
    if (maxpath <= 0) return NULL;
    
    /* check if we can find it now */
    strncpy(name, file, len);
    name[len] = '\0';
    
#  ifdef DEBUG_OPTIONS
    D_OPTIONS2("search_path():  Checking for file \"%s\"\n", name);
#  endif
    if (!access(name, R_OK)) {
	stat(name, &fst);
	if (!S_ISDIR(fst.st_mode)) return name;
    }
    if (ext) {
	strcat(name, ext);
#  ifdef DEBUG_OPTIONS
	D_OPTIONS2("search_path():  Checking for file \"%s\"\n", name);
#  endif
	if (!access(name, R_OK)) {
	    stat(name, &fst);
	    if (!S_ISDIR(fst.st_mode)) return name;
	}
    }
    
    for (path = pathlist; path != NULL && *path != '\0'; path = p) {
	
	int n;
	
	/* colon delimited */
	if ((p = strchr(path, ':')) == NULL) p = strchr(path, '\0');
	
	n = (p - path);
	if (*p != '\0') p++;
	
	if (n > 0 && n <= maxpath) {
	    
	    strncpy(name, path, n);
	    if (name[n-1] != '/') name[n++] = '/';
	    name[n] = '\0';
	    strncat(name, file, len);
	    
#  ifdef DEBUG_OPTIONS
	    D_OPTIONS2("search_path():  Checking for file \"%s\"\n", name);
#  endif
	    if (!access(name, R_OK)) {
		stat(name, &fst);
		if (!S_ISDIR(fst.st_mode)) return name;
	    }
	    if (ext) {
		strcat(name, ext);
#  ifdef DEBUG_OPTIONS
		D_OPTIONS2("search_path():  Checking for file \"%s\"\n", name);
#  endif
		if (!access(name, R_OK)) {
		    stat(name, &fst);
		    if (!S_ISDIR(fst.st_mode)) return name;
		}
	    }
	}
    }
    return NULL;
}
# endif	/* PIXMAP_SUPPORT || (MENUBAR_MAX) */

# ifdef PIXMAP_SUPPORT
/* I think it might be cool to make Eterm load the pixmaps in background.
 * You'd be able to start typing commands without waiting for the bg
 * pixmap processing to end. Thoughts right now: fork(), pthreads. -vendu
 */

void
set_bgPixmap (const char *file)
{
  const char *f;
  static short imlib_init = 0;
  
  /* FIXME: assert() looks like a bad thing. Calls abort(). IMHO, stupid. */
  assert(file != NULL);
  
#  ifdef DEBUG_IMLIB
  fprintf(stderr, "[debug] set_bgPixmap(%s)\n", file);
#  endif
  
  /* Turn on scaling */
  if ((Options & Opt_pixmapScale) || (rs_pixmapScale)) {
	bgPixmap.h = 100; bgPixmap.w = 100;
  }
  
  /*    XSetWindowBackground(Xdisplay, TermWin.vt, PixColors[bgColor]); */
  
  if (*file != '\0') {
	/*	XGetWindowAttributes(Xdisplay, TermWin.vt, &attr); */
	
	/* search environment variables here too */
#  ifdef USE_IMLIB
	if ((f = search_path(rs_path, file, PIXMAP_EXT)) == NULL)
#  endif
#  ifdef PATH_ENV
#   ifdef USE_IMLIB
	  if ((f = search_path(getenv (PATH_ENV), file, PIXMAP_EXT)) == NULL)
#   endif
#  endif /* PATH_ENV */
#  ifdef USE_IMLIB
	    f = search_path(getenv("PATH"), file, PIXMAP_EXT);
#  endif
	
#  ifdef USE_IMLIB
	if (f != NULL) {
	  if (imlib_bg.im != NULL) {
		D_IMLIB("ImlibDestroyImage()\n");
		ImlibDestroyImage(imlib_id, imlib_bg.im);
		imlib_bg.im = NULL;
	  }
	  D_IMLIB2("ReadImageViaImlib(%s)\n", (char *)f);
	  imlib_bg.im = ReadImageViaImlib(Xdisplay, (char *)f);
	}
	
	if (imlib_bg.im == NULL) {
#  endif
	  char * p;
	  
	  if ((p = strchr(file, ';')) == NULL) p = strchr(file, '\0');
	  print_error("couldn't load image file \"%.*s\"", (p - file), file);
#if 0
	  if (!(background_is_pixmap())) {
		D_PIXMAP("set_bgPixmap(): resize_pixmap(TermWin.vt), case 1\n");
		/*	resize_pixmap(TermWin.vt, imlib_bg, bgPixmap, 0, 1); */
		XSetWindowBackground(Xdisplay, TermWin.vt, PixColors[bgColor]);
	  }
#endif
#ifdef POSIX_THREADS
	  XSetWindowBackground(Xdisplay, TermWin.vt, PixColors[bgColor]);
#endif
	} else {
#ifdef POSIX_THREADS
	  D_THREADS("pthread_attr_init(resize_sub_thr_attr)\n");
	  pthread_attr_init(&resize_sub_thr_attr);
# ifdef MUTEX_SYNCH
	  if (pthread_mutex_trylock(&mutex) == EBUSY) {
		D_THREADS("mutex locked, canceling thread\n");
		D_THREADS("set_bgPixmap(): pthread_cancel(resize_sub_thr);\n");
		pthread_cancel(resize_sub_thr);
	  } else {
		D_THREADS("pthread_mutex_trylock(&mutex): ");
		pthread_mutex_unlock(&mutex);
		D_THREADS("pthread_mutex_unlock(&mutex)\n");
	  }
# else
	  if (bg_set) {
		D_THREADS("Background set, cancelling thread\n");
		pthread_cancel(resize_sub_thr);
	  } 
# endif

	  if (!(pthread_create(&resize_sub_thr, &resize_sub_thr_attr,
						   (void *)&resize_bg_thread, NULL))) {
		bg_set = 0;
# ifdef MUTEX_SYNCH
		/*		D_THREADS("pthread_mutex_lock(&mutex);\n"); */
		/*		pthread_mutex_lock(&mutex); */
# endif
		D_THREADS("thread created\n");
	  } else {
		D_THREADS("pthread_create() failed!\n");
	  }
	  
#else
	  if (bg_needs_update) {
		D_PIXMAP("set_bgPixmap(): resize_pixmap(TermWin.vt), case 2\n");
		resize_pixmap(TermWin.vt, imlib_bg, bgPixmap, 0, 1);
		scr_touch();
		bg_needs_update = 0;
	  }
#endif
	}
	
	/*	init_imlib = 1; */
	/*	pthread_join(resize_sub_thr, NULL); */
	D_PIXMAP("set_bgPixmap() exitting\n");
  }
  
  /* FIXME: The call to XClearWindow could probably be removed. -vendu */
  /*    XClearWindow(Xdisplay, TermWin.vt); */
  /*    scr_touch(); */
  /*    XFlush(Xdisplay); */
}

# ifdef USE_IMLIB
#  undef PIXMAP_EXT
# endif

# ifdef PIXMAP_OFFSET
Pixmap
offset_pixmap(Pixmap pmap, int x_off, int y_off)
{
   int pmap_w, pmap_h, bg_w, bg_h;
   int last_xoff=-1, last_yoff=-1;
#  ifdef PIXMAP_SCROLLBAR
    int scr_x, scr_y;
    int scr_w, scr_h;
#  endif
    int root_x, root_y;
    int wide, high;
#  if 0
    int hrz_pix, vrt_pix;    
    int root_w, root_h;
    int dest_x = 0, dest_y = 0;
    int i;
#  endif
    Pixmap tmp;
    XGCValues gcvalue;
    GC gc;
    Window W_dummy;
    int dummy;
    unsigned int u_dummy;
    
    if ((x_off == last_xoff) && (y_off == last_yoff)) {
	return None;
    } else {
	last_xoff = x_off; last_yoff = y_off;
    }
    
    /* Get the width and height of the target pixmap (Eterm background). */
    /* Ignore Xroot, x, y, border width, and depth. */
    XGetGeometry(Xdisplay, TermWin.vt, &W_dummy, &dummy, &dummy,
		 &bg_w, &bg_h, &u_dummy, &u_dummy);
    
#  ifdef DEBUG_PIXMAP
    fprintf(stderr, "offset_pixmap(%d, %d)\n", x_off, y_off);
#  endif
    
    gcvalue.foreground = PixColors [bgColor];
    gc = XCreateGC (Xdisplay, TermWin.vt, GCForeground, &gcvalue);
    
    /* The width and height of the pixmap */
    pmap_w = imlib_bg.im->rgb_width;
    pmap_h = imlib_bg.im->rgb_height;
    
#  ifdef PIXMAP_SCROLLBAR
    /* FIXME: this is a horrible kludge :) -vendu */
    /* The width and height of the scrollbar's background. */
    scr_w = scrollbar_total_width();
    scr_h = bg_h - 1;
#  endif    
    
    /* max offset = width/height of the pixmap */
#  if 0    
    x_off = abs(x_off) % bg_w;
    y_off = abs(y_off) % bg_h;
#  endif
    
    x_off = abs(x_off) % pmap_w;
    y_off = abs(y_off) % pmap_h;
    
#  ifdef DEBUG_PIXMAP
    fprintf(stderr, "XCreatePixmap(%d,%d)\n", bg_w, bg_h);
#  endif
    tmp = XCreatePixmap (Xdisplay, TermWin.vt, bg_w, bg_h, Xdepth);
    
#  ifdef DEBUG_PIXMAP    
    fprintf(stderr,
	    "x_off = %d, y_off = %d, bg_w == %d, bg_h == %d, pmap_w = %d, pmap_h = %d\n",
	    x_off, y_off, bg_w, bg_h, pmap_w, pmap_h);
#  endif
    
    wide = (pmap_w - x_off >= bg_w);
    high = (pmap_h - y_off >= bg_h);    
    
#  if 0    
    root_w = Xroot->width;
    root_h = Xroot->height;
#  endif
    
    if (wide && high) {
#  ifdef DEBUG_PIXMAP
	dprintf(("wide && high\n"));
#  endif
	XTranslateCoordinates(Xdisplay, TermWin.vt, Xroot,
			      0, 0, &root_x, &root_y, &W_dummy);
#  ifdef DEBUG_PIXMAP
	fprintf(stderr, "XCopyArea((%dx%d)@%d,%d -> 0,0)\n",
		 bg_w, bg_h, x_off, y_off);
#  endif	
	XCopyArea(Xdisplay, pmap, tmp, gc,
		  x_off, y_off,
		  bg_w, bg_h,
		  0, 0);
#  ifdef PIXMAP_SCROLLBAR
	if (scrollbar_visible()) {
#   ifdef DEBUG_PIXMAP
	    fprintf(stderr, "XCreatePixmap(sbPixmap.pixmap(%d, %d)\n",
		     scr_w, scr_h);
#   endif	
	    sbPixmap.pixmap = XCreatePixmap(Xdisplay, TermWin.parent,
					    scr_w, scr_h, Xdepth);
	    XTranslateCoordinates(Xdisplay, scrollBar.win, Xroot,
				  0, 0, &scr_x, &scr_y, &W_dummy);
#   ifdef DEBUG_PIXMAP
	    D_PIXMAP("scrollbar: ");
	    fprintf(stderr, "XCopyArea((%dx%d)@%d,%d -> 0,0)\n",
		     scr_w, scr_h, scr_x, scr_y);
#   endif	
	    XCopyArea(Xdisplay, pmap, sbPixmap.pixmap, gc,
		      scr_x, scr_y,
		      scr_w, scr_h,
		      0, 0);
	    D_PIXMAP("XSetWindowBackgroundPixmap(sbPixmap.pixmap)\n");
	    XSetWindowBackgroundPixmap(Xdisplay, scrollBar.win,
				       sbPixmap.pixmap);
	    D_PIXMAP("XFreePixmap(sbPixmap.pixmap)\n");
	    XFreePixmap(Xdisplay, sbPixmap.pixmap);
	    XClearWindow(Xdisplay, scrollBar.win);
	}
#  endif /* PIXMAP_SCROLLBAR */
    } else {
	dprintf(("partial copy\n"));
	XTranslateCoordinates(Xdisplay, TermWin.parent, Xroot,
			      0, 0, &root_x, &root_y, &W_dummy);
	
#  ifdef DEBUG_PIXMAP
	fprintf(stderr, "XCopyArea((%dx%d)@(%d,%d) -> 0,0)\n",
		pmap_w - root_x, pmap_h - root_y, x_off, y_off);
#  endif
	XCopyArea(Xdisplay, pmap, tmp, gc,
		  x_off, y_off,
		  pmap_w - root_x, pmap_h - root_y,
		  0, 0);
#  ifdef PIXMAP_SCROLLBAR
	if (scrollbar_visible()) {
#   ifdef DEBUG_PIXMAP
	    fprintf(stderr, "XCreatePixmap(sbPixmap.pixmap(%d, %d)\n",
		    scr_w, scr_h);
#   endif
	    sbPixmap.pixmap = XCreatePixmap(Xdisplay, TermWin.parent,
					    scr_w, scr_h, Xdepth);
	    XTranslateCoordinates(Xdisplay, scrollBar.win, Xroot,
				  0, 0, &scr_x, &scr_y, &W_dummy);
	    dprintf(("scrollbar: "));
	 
#   ifdef DEBUG_PIXMAP	    
	    fprintf(stderr, "XCopyArea((%dx%d)@%d,%d -> 0,0)\n",
		    scr_w, scr_h, scr_x, scr_y);
#   endif
	      
	    XCopyArea(Xdisplay, pmap, sbPixmap.pixmap, gc,
		      scr_x, scr_y,
		      scr_w, scr_h,
		      0, 0);
	    D_PIXMAP("XSetWindowBackgroundPixmap(sbPixmap.pixmap)\n");
	    XSetWindowBackgroundPixmap(Xdisplay, scrollBar.win,
				       sbPixmap.pixmap);
	    D_PIXMAP("XFreePixmap(sbPixmap.pixmap)\n");
	    XFreePixmap(Xdisplay, sbPixmap.pixmap);	    
	    XClearWindow(Xdisplay, scrollBar.win);
	}
#  endif /* PIXMAP_SCROLLBAR */
    }
    
#  if 0    
    
    /* FIXME: This comment is irrelevant */
    /* Create a temporary Pixmap and copy the Original pixmap with offsets */
    /* into it (4 copies to copy t he four regions). Last copy copies the   */
    /* temporary Pixmap into the Original.                                 */
    
    XCopyArea(Xdisplay, bgPixmap.pixmap, tmp, gc,
	      x_off, y_off,
	      pmap_w - x_off, pmap_h - y_off,
	      0, 0);
    XCopyArea(Xdisplay, bgPixmap.pixmap, tmp, gc,
	      x_off, 0,
	      pmap_w-x_off, y_off,
	      0, pmap_h-y_off);                  
    XCopyArea(Xdisplay, bgPixmap.pixmap, tmp, gc,
	      0, 0,
	      x_off, y_off,
	      pmap_w-x_off, pmap_h-y_off);
    XCopyArea(Xdisplay,bgPixmap.pixmap, tmp, gc,
	      0, y_off,
	      x_off, pmap_h-y_off,
	      pmap_w-x_off, 0);
#  endif
    
    XFreeGC(Xdisplay, gc);
    return tmp;
}
# endif /* PIXMAP_OFFSET) */

#endif	/* PIXMAP_SUPPORT */
