/* threads.c for Eterm
 * Feb 25 1998, vendu
 */

#include <stdio.h>
#include <signal.h>

#include "main.h"
#include "debug.h"
#include "screen.h"

#ifdef PIXMAP_SUPPORT
# include "pixmap.h"
#endif

extern char *sig_to_str(int sig);

#ifdef USE_IMLIB
extern imlib_t imlib_bg;
#endif

#ifdef POSIX_THREADS
# include "threads.h"

extern int refresh_count, refresh_limit;
extern unsigned char * cmdbuf_base, * cmdbuf_ptr, * cmdbuf_endp;
extern unsigned char cmd_getc(void);

# ifdef MUTEX_SYNCH
pthread_mutex_t mutex = PTHREAD_MUTEX_INITIALIZER;
# endif

short bg_set = 0;
pthread_t resize_sub_thr;
pthread_attr_t resize_sub_thr_attr;

# if 0
RETSIGTYPE
Thread_finish(int sig)
{
    D_CMD3("Thread_finish() received signal %s (%d)\n", sig_to_str(sig), sig);
    D_CMD("Installing signal(SIGUSR1, Thread_wait);\n");
    signal(SIGUSR1, Thread_wait);
    scrollbar_show(0);
    scr_touch();
}

RETSIGTYPE
Thread_wait(int sig)
{
    D_CMD3("Thread_wait() received signal %s (%d)\n", sig_to_str(sig), sig);
    D_CMD("Install signal(SIGUSR1, Thread_wait)\n");
    signal(SIGUSR1, Thread_wait);
}

RETSIGTYPE
Thread_exit(int sig)
{
    D_CMD3("Thread_exit() received signal %s (%d)\n", sig_to_str(sig), sig);
    D_CMD("Installing signal(SIGUSR1, Thread_exit)\n");
    signal(SIGUSR1, Thread_exit);
}

RETSIGTYPE
Enter_main_loop(int sig)
{
    D_CMD3("Enter_main_loop() received signal %s (%d)\n", sig_to_str(sig), sig);
    D_CMD("signal(SIGUSR1, SIG_DFL);\n");
/*    signal(SIGUSR1, Thread_wait); */
    signal(SIGUSR1, SIG_DFL);
/*    signal(SIGCHLD, Exit_signal); */
    scrollbar_show(0);
    scr_touch();
    main_loop();
}
# endif

#ifdef MUTEX_SYNCH
void prepare(void *dummy)
{
  D_THREADS("prepare(): pthread_mutex_trylock(&mutex)\n");
  pthread_mutex_trylock(&mutex);
}

void parent(void *dummy)
{
  D_THREADS("parent(): pthread_mutex_trylock(&mutex)\n");
  pthread_mutex_unlock(&mutex);
}

void child(void *dummy)
{
  D_THREADS("child(): pthread_mutex_unlock(&mutex)\n");
  pthread_mutex_unlock(&mutex);
}
#endif

inline void
refresh_termwin(void)
{
    scrollbar_show(0);
    scr_touch();
}

void 
resize_bg_thread(void *dummy)
{
    void * retval = NULL;

	D_THREADS("resize_bg_thread() entered\n");
    
    if (bg_set) {
# ifdef AGGRESSIVE_THREADS
#if 0
	D_THREADS("pthread_cancel(resize_sub_thr);\n");
	pthread_cancel(resize_sub_thr);
#endif
# else
	D_THREADS("Background already set\n");
	pthread_exit(retval);
# endif
    }

#ifdef MUTEX_SYNCH
	if (pthread_mutex_trylock(&mutex) != EBUSY) {
	  D_THREADS("pthread_mutex_trylock(&mutex): ");
	} else {
	  D_THREADS("cannot lock mutex, bailing\n");
	  pthread_exit(retval);
	}
#endif

	D_THREADS("pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, NULL);\n");
    pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, NULL);
    
	D_THREADS("resize_bg_thread(): resize_pixmap()\n");
	resize_pixmap(TermWin.vt, imlib_bg, bgPixmap, 0, 1);

	bg_set = 1;

# ifdef MUTEX_SYNCH
	D_THREADS("pthread_mutex_unlock(&mutex);\n");	
	pthread_mutex_unlock(&mutex);
# endif
	refresh_termwin();
    
    D_THREADS("pthread_exit()\n");
    pthread_exit(retval);
}

/* Read and process output from the application - threaded version. */
inline void
main_loop_thread(void *ignored)
{
  /*   int ch; */
  register int ch;
  
  D_THREADS2("[%d] main_loop_thread() entered\n", getpid());
  D_CMD2("bg_set = %d\n", bg_set);
  do {
	while ((ch = cmd_getc ()) == 0);	/* wait for something */

	if (ch >= ' ' || ch == '\t' || ch == '\n' || ch == '\r')
	  {
		/* Read a text string from the input buffer */
		int nlines = 0;
		/*	   unsigned char * str; */
		register unsigned char * str;	    
		
		/*
		 * point to the start of the string,
		 * decrement first since already did get_com_char ()
		 */
		str = --cmdbuf_ptr;
		while (cmdbuf_ptr < cmdbuf_endp) {
		  ch = *cmdbuf_ptr++;
		  if (ch >= ' ' || ch == '\t' || ch == '\r') {
			/* nothing */
		  } else if (ch == '\n') {
			nlines++;
			if (++refresh_count >= (refresh_limit * (TermWin.nrow - 1)))
			  break;
		  } else {	/* unprintable */
			cmdbuf_ptr--;
			break;
		  }
		}
	      scr_add_lines (str, nlines, (cmdbuf_ptr - str));
	  }
	else
	  {
		switch (ch) {
#ifdef NO_VT100_ANS
		case 005: break;
#else
		case 005: tt_printf (VT100_ANS);	break;	/* terminal Status */
#endif
		case 007: scr_bell ();		break;	/* bell */
		case '\b': scr_backspace ();	break;	/* backspace */
		case 013:
		case 014: scr_index (UP);	break;	/* vertical tab, form feed */
		case 016: scr_charset_choose (1); break;	/* shift out - acs */
		case 017: scr_charset_choose (0); break;	/* shift in - acs */
		case 033: process_escape_seq ();	break;
		}
	  }
  } while (ch != EOF);
  
#if 0
# ifdef MUTEX_SYNCH
  if (pthread_mutex_trylock(&mutex) != EBUSY)
# else
	if (bg_set)
# endif
	  {
# ifdef MUTEX_SYNCH
		D_THREADS("pthreads_mutex_trylock(&mutex)\n");
# endif
		refresh_termwin();
# ifdef MUTEX_SYNCH
		D_THREADS("cmd_getc(): pthread_mutex_unlock(&mutex);\n");    
		pthread_mutex_unlock(&mutex);
# endif
		bg_set = 0;
	  }
#endif
  
  
#  if 0
  if (pthread_mutex_trylock(&mutex) == EBUSY) {
	D_THREADS("main_loop_thread(): mutex locked, bbl\n");
	main_loop_thread(NULL);
  } else {
	D_THREADS("main_loop_thread(): pthread_mutex_lock(&mutex);\n");
	/*	  pthread_mutex_lock(&mutex); */
	if (bg_set) refresh_termwin();
	D_THREADS("main_loop_thread(): pthread_mutex_unlock(&mutex);\n");    
	pthread_mutex_unlock(&mutex);
	bg_set = 0;
	main_loop_thread(NULL);
  }
# endif
}

# if 0
void start_bg_thread(void)
{
    static int create_val = -1;
    void *exit_val;
    static pthread_t bg_pixmap_thr;
    static pthread_attr_t bg_thr_attr;
    
    D_THREADS("start_bg_thread: ");
    pthread_attr_destroy(&bg_thr_attr);    
    pthread_attr_init(&bg_thr_attr);
    D_THREADS("initialized attributes, ");
    pthread_attr_setdetachstate(&bg_thr_attr, PTHREAD_CREATE_DETACHED);
    D_THREADS("set to be detached, ");
    pthread_cancel(bg_pixmap_thr);
    D_THREADS("canceled.");
	/*    signal(SIGUSR1, Thread_wait); */
    if (!(create_val =
	  pthread_create(&bg_pixmap_thr, &bg_thr_attr,
			 (void *)&resize_bg_thread, NULL))) {
	/* NOTREACHED */
	D_THREADS("pthread_create(): ");
	return;
    } else if (create_val == EAGAIN) {
	D_THREADS("failed!\n");
    }
}
# endif

#endif /* POSIX_THREADS */
